// ===================================================================
// $Id: $
//
// OBJloader.cpp
//     Source file for Wawefront OBJ file geometry loading
//
// Initial coding by Jiri Filip, 2004.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <cassert>
#include "TAlloc.h"
#include "OBJloader.h"



//########################################################################################
//#   TObjectData                                                                        #
//########################################################################################

TObjectData::TObjectData(int dataSize)
{
  this->dataSize = dataSize;
  fileName = callocation1(0,1000);
  OBJfileName = callocation1(0,1000);
  vertices = allocation2(1,dataSize,0,3);
  UVcoord = allocation2(1,dataSize,0,3);
  normals = allocation2(1,dataSize,0,3);
  faces = lallocation2(1,dataSize,0,8);

  tangentsF = NULL; 
  tangentsV = NULL; 
  normalsF = NULL; 
  normalsV = NULL; 
  binormalsV = NULL; 
  binormalsF = NULL; 
}//--- TObjectData --------------------------------------------------------

TObjectData::~TObjectData()
{
  freecmem1(OBJfileName, 0,1000);
  freecmem1(fileName, 0,1000);  

  freemem2(vertices,1,dataSize,0,3);
  freemem2(UVcoord,1,dataSize,0,3);
  freemem2(normals,1,dataSize,0,3);
  freelmem2(faces,1,dataSize,0,8);

  if(tangentsF!=NULL)
    freemem2(tangentsF,1,dataSize,0,2); 
  if(tangentsV!=NULL)
    freemem2(tangentsV,1,dataSize,0,2); 
  if(normalsF!=NULL)
    freemem2(normalsF,1,dataSize,0,2); 
  if(normalsV!=NULL)
    freemem2(normalsV,1,dataSize,0,2); 
  if(binormalsV!=NULL)
    freemem2(binormalsV,1,dataSize,0,2); 
  if(binormalsF!=NULL)
  freemem2(binormalsF,1,dataSize,0,2); 
}//--- TObjectData --------------------------------------------------------

int TObjectData::getNoOfTokens(char *line, char *delimiter)
{
  char *token;

  int NoOfTokens=0;
  token = strtok(line, delimiter);
  while(token!=NULL)
    {
      //      printf("%s\n",token);
      NoOfTokens++;
      token = strtok(NULL,delimiter);
    }
  return NoOfTokens;
}//--- getNoOfTokens -----------------------------------------------------

int TObjectData::loadOBJfile(char *_fileName, int center, float scale, float xoff, float yoff, float zoff)
{
  strcpy(OBJfileName, _fileName);

  printf("Loading 3D geometry from file '%s'...",_fileName);fflush(NULL);
  FILE *fp;

  if((fp=fopen(_fileName,"r"))==NULL) {
    printf("Filename '%s' is wrong or file doesn't exist\n",_fileName);
    exit(0);
  }

  i_v = 1; i_vt = 1; i_vn = 1; i_f = 1;
  char dataType[100];
  float v[3], vt[3];
  int f[3], ft[3];
  int assignedItems;
  int NoOfTokens;
  char * line = NULL;
  size_t len = 0;
  ssize_t read;
  char dt[100];
  char line2[1000];
  int lineIndex=0;
  long int polygon[4][3];

  while ((read = getline(&line, &len, fp)) != -1) {
    lineIndex++;
    //    printf("Retrieved line of length %zu :\n", read);
    assignedItems = sscanf(line,"%s", (char*)(&dataType));
    //printf("%s %s %d\n", line, dataType, assignedItems);
    strcpy(line2,line);
    NoOfTokens = getNoOfTokens(line, (char*)" ");
    //    printf("%d\n",NoOfTokens);
    //    exit(0);

    if(assignedItems == 1) { // if not empty line
      if (strcmp(dataType,"v")==0)
	if (NoOfTokens==4) {
	  sscanf(line2,"%s %f %f %f", (char*)&dt, &vertices[i_v][0], &vertices[i_v][1], &vertices[i_v][2]);
	  i_v++;
	}
	else {
	  printf("  Line: %d - Wrong number of vertex coordinates %d (expected 3)!!! Exiting...\n",
		 lineIndex, NoOfTokens-1);
	  exit(-1);
	}

      if (strcmp(dataType,"vt")==0) {
	if (NoOfTokens==3) {
	  sscanf(line2,"%s %f %f\n",(char*)&dt, &UVcoord[i_vt][0],&UVcoord[i_vt][1]);
	  i_vt++;
	}
	else if(NoOfTokens==4) {
	  sscanf(line2,"%s %f %f %f\n",(char*)&dt, &UVcoord[i_vt][0],&UVcoord[i_vt][1],&UVcoord[i_vt][2]);
	  i_vt++;
	}
	else {
	  printf("  Line: %d - Wrong number of UV coordinates %d (expected 2 or 3)!!! Exiting...\n",lineIndex,NoOfTokens-1);
	  exit(-1);
	}
      }
      if (strcmp(dataType,"vn")==0)
	if (NoOfTokens==4) {
	  sscanf(line2,"%s %f %f %f\n", (char*)&dt, &normals[i_vn][0],&normals[i_vn][1],&normals[i_vn][2]);
	  i_vn++;
	}
	else {
	  printf("  Line: %d - Wrong number of normal coordinates %d (expected: 3)!!! Exiting...\n",lineIndex,NoOfTokens-1);
	  exit(-1);
	}

        if (strcmp(dataType,"f")==0) {
	  if (NoOfTokens==4) {
	    if((i_vn>1)&&(i_vt>1))
	      sscanf(line2,"%s %ld/%ld/%ld %ld/%ld/%ld %ld/%ld/%ld\n",(char*)&dt, &faces[i_f][0],
		     &faces[i_f][3],&faces[i_f][6],&faces[i_f][1],&faces[i_f][4],
		     &faces[i_f][7],&faces[i_f][2],&faces[i_f][5],&faces[i_f][8]);
	    // f v//vn
	    if ((i_vn>1)&&(i_vt==1))
	      sscanf(line2,"%s %ld//%ld %ld//%ld %ld//%ld\n",(char*)&dt, &faces[i_f][0],&faces[i_f][6],
		     &faces[i_f][1],&faces[i_f][7],&faces[i_f][2],&faces[i_f][8]);
	    
	    // f v/vt
	    if ((i_vn==1)&&(i_vt>1))
	      sscanf(line2,"%s %ld/%ld %ld/%ld %ld/%ld\n",(char*)&dt, &faces[i_f][0],&faces[i_f][3],
		     &faces[i_f][1],&faces[i_f][4],&faces[i_f][2],&faces[i_f][5]);
	    i_f++;
	  }
	  else if(NoOfTokens==5) {
	    if ((i_vn>1)&&(i_vt>1)) {
	      sscanf(line2,"%s %ld/%ld/%ld %ld/%ld/%ld %ld/%ld/%ld %ld/%ld/%ld\n",(char*)&dt, &polygon[0][0],
		     &polygon[0][1],&polygon[0][2],  &polygon[1][0],&polygon[1][1],&polygon[1][2],
		     &polygon[2][0],&polygon[2][1],&polygon[2][2],  &polygon[3][0],&polygon[3][1],&polygon[3][2]);
	      // first triangle
	      // vertices
	      faces[i_f][0] = polygon[0][0];
	      faces[i_f][1] = polygon[1][0];
	      faces[i_f][2] = polygon[2][0];
	      // UV coords.
	      faces[i_f][3] = polygon[0][1];
	      faces[i_f][4] = polygon[1][1];
	      faces[i_f][5] = polygon[2][1];
	      // normals
	      faces[i_f][6] = polygon[0][2];
	      faces[i_f][7] = polygon[1][2];
	      faces[i_f][8] = polygon[2][2];
	      i_f++;
	      // second triangle
	      // vertices
	      faces[i_f][0] = polygon[3][0];
	      faces[i_f][1] = polygon[0][0];
	      faces[i_f][2] = polygon[2][0];
	      // UV coords.
	      faces[i_f][3] = polygon[3][1];
	      faces[i_f][4] = polygon[0][1];
	      faces[i_f][5] = polygon[2][1];
	      // normals
	      faces[i_f][6] = polygon[3][2];
	      faces[i_f][7] = polygon[0][2];
	      faces[i_f][8] = polygon[2][2];
	      i_f++;                    
	    }
            
	    // f v//vn
	    if ((i_vn>1)&&(i_vt==1)) {
	      sscanf(line2,"%s %ld//%ld %ld//%ld %ld//%ld %ld//%ld\n",(char*)&dt, &polygon[0][0],
		     &polygon[0][2],&polygon[1][0],&polygon[1][2],&polygon[2][0],
		     &polygon[2][2],&polygon[3][0],&polygon[3][2]);
	      // first triangle
	      // vertices
	      faces[i_f][0] = polygon[0][0];
	      faces[i_f][1] = polygon[1][0];
	      faces[i_f][2] = polygon[2][0];
	      // normals
	      faces[i_f][6] = polygon[0][2];
	      faces[i_f][7] = polygon[1][2];
	      faces[i_f][8] = polygon[2][2];
	      i_f++;
	      // second triangle
	      // vertices
	      faces[i_f][0] = polygon[3][0];
	      faces[i_f][1] = polygon[0][0];
	      faces[i_f][2] = polygon[2][0];
	      // normals
	      faces[i_f][6] = polygon[3][2];
	      faces[i_f][7] = polygon[0][2];
	      faces[i_f][8] = polygon[2][2];
	      i_f++;                    
	    }
	    // f v/vt
	    if ((i_vn==1)&&(i_vt>1)) {
	      sscanf(line2,"%s %ld/%ld %ld/%ld %ld/%ld %ld/%ld\n",(char*)&dt, &polygon[0][0],&polygon[0][1],
		     &polygon[1][0],&polygon[1][1],&polygon[2][0],&polygon[2][1],&polygon[3][0],&polygon[3][1]);
	      // first triangle
	      // vertices
	      faces[i_f][0] = polygon[0][0];
	      faces[i_f][1] = polygon[1][0];
	      faces[i_f][2] = polygon[2][0];
	      // UV coords.
	      faces[i_f][3] = polygon[0][1];
	      faces[i_f][4] = polygon[1][1];
	      faces[i_f][5] = polygon[2][1];
	      i_f++;
	      // second triangle
	      // vertices
	      faces[i_f][0] = polygon[3][0];
	      faces[i_f][1] = polygon[0][0];
	      faces[i_f][2] = polygon[2][0];
	      // UV coords.
	      faces[i_f][3] = polygon[3][1];
	      faces[i_f][4] = polygon[0][1];
	      faces[i_f][5] = polygon[2][1];
	      i_f++;                    
	    }
	  }
	  else {
	    printf("  Line: %d - Wrong number of polygon vertices %d (expected: 3 or 4)!!! Exiting...\n",lineIndex,NoOfTokens-1);
	    exit(-1);
	  }	  
	}// f
    }
  } // while not final line

  printf("done\n");
  printf("   Total of:\n        %10d vertices\n        %10d UV coordinates\n        %10d normals\n"
	 "        %10d faces               were loaded\n\n", (int)i_v-1,(int)i_vt-1,(int)i_vn-1,(int)i_f-1);
  fclose(fp);
  // vertices centering and scaling
  long i;
  double sum[3];
  if(center)
    {
      sum[0] = sum[1] = sum[2] = 0.0;
      for(i=1;i<i_v;i++)
        {
          sum[0] += vertices[i][0];
          sum[1] += vertices[i][1];
          sum[2] += vertices[i][2];
        }
      for(i=0;i<3;i++)
        sum[i] /= i_v-1;
      printf("   Object Center: %3.2f %3.2f %3.2f\n",sum[0],sum[1],sum[2]);
      for(i=1;i<i_v;i++)
        {
          vertices[i][0] -= sum[0];
          vertices[i][1] -= sum[1];
          vertices[i][2] -= sum[2];
        }
    }

  for(i=1;i<i_v;i++)
    {
      vertices[i][0] -= xoff;
      vertices[i][1] -= yoff;
      vertices[i][2] -= zoff;
    }

  // scaling vertices
  for(i=1;i<i_v;i++)
    {
      vertices[i][0] *= scale;
      vertices[i][1] *= scale;
      vertices[i][2] *= scale;
    }

  // checking UV coords if correction needed
  float Xmax=-1e10,Xmin=1e10;
  float Ymax=-1e10,Ymin=1e10;
  for(i=1;i<i_vt;i++)
    {
      if(UVcoord[i][0]>Xmax)
        Xmax = UVcoord[i][0];
      if(UVcoord[i][0]<Xmin)
        Xmin = UVcoord[i][0];
      if(UVcoord[i][1]>Ymax)
        Ymax = UVcoord[i][1];
      if(UVcoord[i][1]<Ymin)
        Ymin = UVcoord[i][1];
    }
  for(i=1;i<i_vt;i++)
    {
      UVcoord[i][0] = (UVcoord[i][0]-Xmin)/(Xmax-Xmin); 
      UVcoord[i][1] = (UVcoord[i][1]-Ymin)/(Ymax-Ymin); 
    }

  // nomals normalization
  for(i=1;i<i_vn;i++)
    vec_norm(normals[i]);

  return 1;
}//--- loadOBJfile -----------------------------------------------------------------

void TObjectData::normalsCorrection()
{
  //chech CCW ordering of vertices in the triangle
  int i,j,k,l, ii,jj,kk,ll;
  long A,B,C, AA,BB,CC;
  int count = 0;
  for(i=1;i<i_f;i++)
    for(j=0;j<=2;j++) {
      if(j==2)
	k = 0;
      else
	k = j+1;
      if(k==2)
	l = 0;
      else
	l = k+1;

      A = faces[i][j];
      B = faces[i][k];
      C = faces[i][l];
      for(ii=1;ii<i_f;ii++)
	for(jj=0;jj<=2;jj++) {
	  if(jj==2)
	    kk = 0;
	  else
	    kk = jj+1;
	  if(kk==2)
	    ll = 0;
	  else
	    ll = kk+1;
              
	  AA = faces[ii][jj];
	  BB = faces[ii][kk];
	  CC = faces[ii][ll];
	  // if CCW ordering invert normals
	  if ((A==AA)&&(B==BB)&&(C!=CC)) {
	    printf("%d %d %d   %d %d %d     %d %d %d   %d %d %d\n",
		   (int)A, (int)B, (int)C, (int)j, (int)k, (int)l,
		   (int)AA, (int)BB, (int)CC, (int)jj, (int)kk, (int)ll);
	    count++;
	    // v1
	    normals[faces[ii][6]][0] *= -1;
	    normals[faces[ii][6]][1] *= -1;
	    normals[faces[ii][6]][2] *= -1;
	    // v2
	    normals[faces[ii][7]][0] *= -1;
	    normals[faces[ii][7]][1] *= -1;
	    normals[faces[ii][7]][2] *= -1;
	    // v3
	    normals[faces[ii][8]][0] *= -1;
	    normals[faces[ii][8]][1] *= -1;
	    normals[faces[ii][8]][2] *= -1;
	    //                  jj = 3; // end the closest for(jj)
	  }
	}
    }
  printf("Corrected normals: %d\n",count);

}//--- normalsCorrection -----------------------------------------------------------

void TObjectData::saveLoadedOBJ(char *fileName, int off_v, int off_vt, int off_vn)
{
  printf("Saving 3D geometry from file '%s'...\n",fileName);

  int i;
  FILE *fp;
  if((fp=fopen(fileName,"w"))==NULL)
    {
      printf("Filename '%s' is wrong or file doesn't exist\n",fileName);
      exit(0);
    }
  
  fprintf(fp,"# OBJ file (generated using PR UTIA parser)\n\n");

  for(i=1;i<i_v;i++)
    fprintf(fp,"v %f %f %f\n",vertices[i][0],vertices[i][1],vertices[i][2]);
  fprintf(fp,"\n");
  for(i=1;i<i_vt;i++)
    fprintf(fp,"vt %f %f %f\n",UVcoord[i][0],UVcoord[i][1],UVcoord[i][2]);
  fprintf(fp,"\n");
  if(i_vn>1)
    {
      for(i=1;i<i_vn;i++)
        fprintf(fp,"vn %f %f %f\n",normals[i][0],normals[i][1],normals[i][2]);
      fprintf(fp,"\n");
      for(i=1;i<i_f;i++)
        fprintf(fp,"f %ld/%ld/%ld %ld/%ld/%ld %ld/%ld/%ld\n",faces[i][0]-off_v,faces[i][3]-off_vt,faces[i][6]-off_vn,faces[i][1]-off_v,faces[i][4]-off_vt,faces[i][7]-off_vn,faces[i][2]-off_v,faces[i][5]-off_vt,faces[i][8]-off_vn);
    }
  else
    for(i=1;i<i_f;i++)
      fprintf(fp,"f %ld/%ld %ld/%ld %ld/%ld\n",faces[i][0]-off_v,faces[i][3]-off_vt,faces[i][1]-off_v,faces[i][4]-off_vt,faces[i][2]-off_v,faces[i][5]-off_vt);

  fclose(fp);
}//--- saveLoadedOBJ -------------------------------------------------------------------

void TObjectData::splitOBJ(char *fileName)
{
  printf("Splitting 3D geometry from file '%s'...\n",fileName);

  char name[1000];
  strcpy(name,"");
  strncpy(name,fileName,strlen(fileName)-4);
  name[strlen(fileName)-4]='\0';

  FILE *fp;

  if((fp=fopen(fileName,"r"))==NULL)
    {
      printf("Filename '%s' is wrong or file doesn't exist\n",fileName);
      exit(0);
    }

  char dataType[1000];
  float v[3], vt[3];
  int f[3], ft[3];

  int groupNo=0;
  int last_v=0,last_vt=0,last_vn=0;
  char files[100];

  i_v=1; i_vt=1; i_vn=1; i_f=1;
  do {
    int tt = fscanf(fp,"%s\n",(char*)(&dataType)); 
    if(strcmp(dataType,"v")==0) {
      tt = fscanf(fp,"%f %f %f\n",&vertices[i_v][0],&vertices[i_v][1],&vertices[i_v][2]);
      //          printf("v %3.5f %3.5f %3.5f\n",vertices[i_v][0],vertices[i_v][1],vertices[i_v][2]);
      i_v++;
    }
    if(strcmp(dataType,"vt")==0) {
      tt = fscanf(fp,"%f %f %f\n",&UVcoord[i_vt][0],&UVcoord[i_vt][1],&UVcoord[i_vt][2]);
      //          printf("vt %3.5f %3.5f %3.5f\n",UVcoord[i_vt][0],UVcoord[i_vt][1],UVcoord[i_vt][2]);
      i_vt++;
    }
    if(strcmp(dataType,"vn")==0) {
      tt = fscanf(fp,"%f %f %f\n",&normals[i_vn][0],&normals[i_vn][1],&normals[i_vn][2]);
      //          printf("vn %3.5f %3.5f %3.5f\n",normals[i_vn][0],normals[i_vn][1],normals[i_vn][2]);
      i_vn++;
    }      
    if (strcmp(dataType,"f")==0) {
      if (i_vn > 1) {
	tt = fscanf(fp,"%ld/%ld/%ld %ld/%ld/%ld %ld/%ld/%ld\n",
		    &faces[i_f][0],&faces[i_f][3],&faces[i_f][6],&faces[i_f][1],
		    &faces[i_f][4],&faces[i_f][7],&faces[i_f][2],&faces[i_f][5],&faces[i_f][8]);
	// printf("f %ld/%ld %ld/%ld %ld/%ld       %ld\n",faces[i_f][0],faces[i_f][3],faces[i_f][1],faces[i_f][4],faces[i_f][2],faces[i_f][5],   i_f);
      }
      else {
	tt = fscanf(fp,"%ld/%ld %ld/%ld %ld/%ld\n",&faces[i_f][0],&faces[i_f][3],&faces[i_f][1],
		    &faces[i_f][4],&faces[i_f][2],&faces[i_f][5]);
	// printf("f %ld/%ld %ld/%ld %ld/%ld       %ld\n",faces[i_f][0],faces[i_f][3],faces[i_f][1],faces[i_f][4],faces[i_f][2],faces[i_f][5],   i_f);
      }
      i_f++;
    }

    if((strcmp(dataType,"g")==0)&&(i_f>1)) {
      printf(" Saving Group: %d\n",groupNo);
      printf("Total of:\n     %10d vertices\n     %10d UV coordinates\n     %10d normals\n     %10d faces"
	     "               were loaded\n\n", (int)i_v-1, (int)i_vt-1, (int)i_vn-1, (int)i_f-1);

      // saving submodel
      sprintf(files,"%s%d.obj",name,groupNo);
      saveLoadedOBJ(files, last_v,last_vt,last_vn);

      last_v += i_v-1;
      last_vt += i_vt-1;
      last_vn += i_vn-1;
      printf("offset: %d %d %d\n",last_v,last_vt,last_vn);
      
      i_v=1; i_vt=1; i_vn=1; i_f=1;
      groupNo++;
    }
  }
  while(feof(fp)==0);
  printf(" Saving Group: %d\n",groupNo);
  printf("Total of:\n     %10d vertices\n     %10d UV coordinates\n     %10d normals\n     %10d faces"
	 "               were loaded\n\n", (int)i_v-1, (int)i_vt-1, (int)i_vn-1, (int)i_f-1);

  // saving submodel
  //  sprintf(files,"submodel%d.obj",groupNo);
  sprintf(files,"%s%d.obj",name,groupNo);
  saveLoadedOBJ(files, last_v,last_vt,last_vn);

  fclose(fp);  
}//--- splitOBJ ------------------------------------------------------------------------

long TObjectData::noOfFaces()
{
  return (i_f-1);
}//--- noOfFaces -----------------------------------------------------------------------

long TObjectData::noOfVertices()
{
  return (i_vt-1);
}//--- noOfVertices -----------------------------------------------------------------------

long TObjectData::noOfNormals()
{
  return (i_vn-1);
}//--- noOfNormalss -----------------------------------------------------------------------

void TObjectData::testOnDegeneratedPolygons()
{
  printf("Testing degenerated polygons...      \r");fflush(stdout);
  int flag1 = 0;
  for(int k=1;k<i_f;k++)
    {
      if( (vertices[faces[k][0]][0] == vertices[faces[k][1]][0] &&
           vertices[faces[k][0]][1] == vertices[faces[k][1]][1] &&
           vertices[faces[k][0]][2] == vertices[faces[k][1]][2] ) ||
          (vertices[faces[k][0]][0] == vertices[faces[k][2]][0] &&
           vertices[faces[k][0]][1] == vertices[faces[k][2]][1] &&
           vertices[faces[k][0]][2] == vertices[faces[k][2]][2] ) ||
          (vertices[faces[k][1]][0] == vertices[faces[k][2]][0] &&
           vertices[faces[k][1]][1] == vertices[faces[k][2]][1] &&
           vertices[faces[k][1]][2] == vertices[faces[k][2]][2] ) )
        {
          flag1++;
          if(flag1==1)
              printf("\nThe same 3D coordinates have polygons: %d, ",k);
          if(flag1>1)
            printf("%d, ",k);
        }
    }

  int flag2 = 0;
  for(int k=1;k<i_f;k++)
    {
      if( faces[k][0] == faces[k][1] || faces[k][0] == faces[k][2] || faces[k][1] == faces[k][2] )
        {
          flag2++;
          if(flag2==1)
              printf("\nIdentical vertices in polygons: %d, ",k);
          if(flag2>1)
            printf("%d, ",k);
        }
    }

  int flag3 = 0;
  float eps=1e-10;
  for(int k=1;k<i_f;k++)
    {
      if( (fabs(UVcoord[faces[k][3]][0] - UVcoord[faces[k][4]][0])<eps &&
           fabs(UVcoord[faces[k][3]][1] - UVcoord[faces[k][4]][1])<eps) ||
          (fabs(UVcoord[faces[k][3]][0] - UVcoord[faces[k][5]][0])<eps &&
           fabs(UVcoord[faces[k][3]][1] - UVcoord[faces[k][5]][1])<eps) ||
          (fabs(UVcoord[faces[k][4]][0] - UVcoord[faces[k][5]][0])<eps &&
           fabs(UVcoord[faces[k][4]][1] - UVcoord[faces[k][5]][1])<eps) )
        {
          flag3++;
          if(flag3==1)
            printf("\nThe same UV coords in polygons: %d, ",k);
          if(flag3>1)
            printf("%d, ",k);
        }
    }

  int flag4 = 0;
  for(int k=1;k<i_f;k++)
    {
      if( (UVcoord[faces[k][3]][0] < 0.f) || 
          (UVcoord[faces[k][4]][0] < 0.f) || 
          (UVcoord[faces[k][5]][0] < 0.f) ||
          (UVcoord[faces[k][3]][1] < 0.f) ||
          (UVcoord[faces[k][4]][1] < 0.f) || 
          (UVcoord[faces[k][5]][1] < 0.f))
        {
          flag4++;
          if(flag4==1)
              printf("\nNegative UV coords in polygons: %d, ",k);
          if(flag4>1)
            printf("%d, ",k);
        }
    }

  if((flag1>=1) || (flag2>=1) || (flag3>=1) || (flag4>=1))
    {
      if(flag1>=1) {
        printf("\nTotal of %d degenerated vertices in the model...done\n",flag1);fflush(stdout); }
      if(flag2>=1) {
        printf("\nTotal of %d degenerated polygons in the model...done\n",flag2);fflush(stdout); }
      if(flag3>=1) {
        printf("\nTotal of %d  degenerated UVcoords in the model...done\n",flag3);fflush(stdout); }
      if(flag4>=1) {
        printf("\nTotal of %d  negative UVcoords in the model...done\n",flag4);fflush(stdout); }
    }
  else
    {printf("Testing degenerated polygons...done     \n");fflush(stdout);}
  return;
}//--- testOnDegeneratedPolygons ---------------------------------------------------------

void TObjectData::testAndCorrectNormalsContinuity()
{
  printf("Testing and correcting of normals continuity...      \r");fflush(stdout);
  if(normalsV==NULL)
    normalsV = allocation2(1,dataSize,0,2); 
  else
    return;
  // computing tangents for individual vertices
  int trianglesPerVertex;
  float v[3][3];

  for(int i=1;i<i_v;i++)
    {
      for(int j=0;j<=2;j++)
        normalsV[i][j] = 0.f;
      trianglesPerVertex = 0;
      for(int k=1;k<i_f;k++)
        {
          for(int j=0;j<=2;j++)
            {
              v[0][j] = vertices[faces[k][0]][j];
              v[1][j] = vertices[faces[k][1]][j];
              v[2][j] = vertices[faces[k][2]][j];
            }
          if( ((v[0][0] == vertices[i][0])&&(v[0][1] == vertices[i][1])&&(v[0][2] == vertices[i][2]))||
              ((v[1][0] == vertices[i][0])&&(v[1][1] == vertices[i][1])&&(v[1][2] == vertices[i][2]))||
              ((v[2][0] == vertices[i][0])&&(v[2][1] == vertices[i][1])&&(v[2][2] == vertices[i][2])) )
            {
              for(int j=0;j<=2;j++)
                normalsV[i][j] += normals[faces[k][0]][j];
              trianglesPerVertex++;
            }
        }// for faces
      if(trianglesPerVertex >= 1)
        for(int j=0;j<=2;j++)
          normalsV[i][j] /= trianglesPerVertex;
      else
        for(int j=0;j<=2;j++)
          normalsV[i][j] = 0.1f;

      vec_norm(normalsV[i]);
    }// for vertices
  printf("Testing and correcting of normals continuity...done   \n");fflush(stdout);
}//--- testAndCorrectNormalsContinuity ---------------------------------------------------------

void TObjectData::normalsPerFace()
{
  printf("Computing normals for individual faces...\r");fflush(stdout);
  if(normalsF==NULL)
    normalsF = allocation2(1,dataSize,0,2); 
  else
    return;
  float **v = allocation2(0,2, 0,2); 
  float **n = allocation2(0,2, 0,2); 
  for(int i=1;i<i_f;i++)
    {
      for(int j=0;j<=2;j++)
        {
          v[0][j] = vertices[faces[i][0]][j];
          v[1][j] = vertices[faces[i][1]][j];
          v[2][j] = vertices[faces[i][2]][j];
        }
      normal(v[0], v[1], v[2], n[1]);
      normal(v[2], v[0], v[1], n[0]);
      normal(v[1], v[2], v[0], n[2]);
      vec_norm(n[0]);
      vec_norm(n[1]);
      vec_norm(n[2]);

      for(int j=0;j<=2;j++)
        normalsF[i][j] =  (n[0][j] + n[1][j] + n[2][j])/3.f;
      vec_norm(normalsF[i]);

    }// for faces
  printf("Computing normals for individual faces...done\n");fflush(stdout);
  freemem2(v,  0,2, 0,2); 
  freemem2(n,  0,2, 0,2); 
  return;
}//--- normalsPerFace ---------------------------------------------------------------------------

void TObjectData::normalsPerVertex()
{
  printf("Computing normals for individual vertices...       \r");fflush(stdout);
  if(normalsV==NULL)
    normalsV = allocation2(1,dataSize,0,2); 
  else
    return;
  if(normalsF==NULL)
    this->normalsPerFace();
  // computing normals for individual vertices
  int trianglesPerVertex;
  for(int i=1;i<i_v;i++)
    {
      for(int j=0;j<=2;j++)
        normalsV[i][j] = 0.f;
      trianglesPerVertex = 0;
      for(int k=1;k<i_f;k++)
        {
          if((i == faces[k][0])||(i == faces[k][1])||(i == faces[k][2]))
            {
              for(int j=0;j<=2;j++)
                normalsV[i][j] += normalsF[k][j];
              trianglesPerVertex++;
            }
        }// for faces
      if(trianglesPerVertex >= 1)
        for(int j=0;j<=2;j++)
          normalsV[i][j] /= trianglesPerVertex;
      else
        for(int j=0;j<=2;j++)
          normalsV[i][j] = 0.1f;

      vec_norm(normalsV[i]);
    }// for vertices
  printf("Computing normals for individual vertices...done\n");fflush(stdout);
  return;
}//--- normalsPerVertex ----------------------------------------------------------------

void TObjectData::computeTB(float **v, float **UV, float *T, float *B)
{
  // compute the 2.0 * surface area of the triangle
  float twoA = fabs( (UV[1][0] - UV[0][0]) * (UV[2][1] - UV[0][1]) +
                     (UV[2][0] - UV[0][0]) * (UV[1][1] - UV[0][1]) );
  //  assert(fabs(twoA) > 1e-12);
  if(twoA < 1e-12)
    twoA = 1e-12;
  twoA = 1.0 / twoA;
  for(int j=0;j<=2;j++)
    {
      T[j] = twoA * (v[0][j] * (UV[1][1] - UV[2][1]) + v[1][j] * (UV[2][1] - UV[0][1]) + v[2][j] * (UV[0][1] - UV[1][1]));
      B[j] = twoA * (v[0][j] * (UV[2][0] - UV[1][0]) + v[1][j] * (UV[0][0] - UV[2][0]) + v[2][j] * (UV[1][0] - UV[0][0]));
    }
  vec_norm(T);
  vec_norm(B);
  return;
}//--- computeTB ---------------------------------------------------

void TObjectData::tangentsPerFace()
{
  if(i_vt==1)
    {
      printf("OBJloader: !!! UV coordinates are not available in the model !!! Exiting...\n");
      exit(-1);
    }
  printf("Computing tangents for individual faces...\r");fflush(stdout);
  if(tangentsF==NULL)
    tangentsF = allocation2(1,dataSize,0,2); 
  else
    return;

  if(binormalsF==NULL)
    binormalsF = allocation2(1,dataSize,0,2); 
  else
    return;

  float den;
  float **UV = allocation2(0,2, 0,2); 
  float **v = allocation2(0,2, 0,2); 
  float **n = allocation2(0,2, 0,2); 

  // computing tangents for individual faces
  for(int i=1;i<i_f;i++)
    {
      //for degenerated triangles skip computations
      if(faces[i][0]==faces[i][1] || faces[i][0]==faces[i][2] || faces[i][2]==faces[i][1])
        {
          for(int j=0;j<=2;j++)
            tangentsF[i][j] = 0.f;
          continue;
        }
      for(int j=0;j<=2;j++)
        {
          v[0][j] = vertices[faces[i][0]][j];
          v[1][j] = vertices[faces[i][1]][j];
          v[2][j] = vertices[faces[i][2]][j];
        }        

      UV[0][0] = UVcoord[faces[i][3]][0];
      UV[0][1] = UVcoord[faces[i][3]][1];
      UV[1][0] = UVcoord[faces[i][4]][0];
      UV[1][1] = UVcoord[faces[i][4]][1];
      UV[2][0] = UVcoord[faces[i][5]][0];
      UV[2][1] = UVcoord[faces[i][5]][1];

      //      computeTB(v, UV, tangentsF[i], binormalsF[i]);

      den = UV[2][0]*(UV[1][1]-UV[0][1]) 
          - UV[1][0]*(UV[2][1]-UV[0][1]) 
          - UV[0][0]*(UV[1][1]-UV[2][1]);  
      //      assert(fabs(den)>1e-15);
      if(fabs(den)>1e-15) {
      tangentsF[i][0] = (v[2][0]*(UV[1][1]-UV[0][1]) 
                       - v[1][0]*(UV[2][1]-UV[0][1]) 
                       - v[0][0]*(UV[1][1]-UV[2][1])) / den;
      tangentsF[i][1] = (v[2][1]*(UV[1][1]-UV[0][1]) 
                       - v[1][1]*(UV[2][1]-UV[0][1]) 
                       - v[0][1]*(UV[1][1]-UV[2][1])) / den;
      tangentsF[i][2] = (v[2][2]*(UV[1][1]-UV[0][1]) 
                       - v[1][2]*(UV[2][1]-UV[0][1]) 
                       - v[0][2]*(UV[1][1]-UV[2][1])) / den;
      vec_norm(tangentsF[i]);
      }
      else
        for(int j=0;j<=2;j++)
          tangentsF[i][j] = 0.1f;

      if(vec_len(tangentsF[i])<=0.f) 
        {
#define HACK
#ifdef HACK
	  tangentsF[i][0] = 1.0;
	  tangentsF[i][1] = 0.0;
	  tangentsF[i][2] = 0.0;
#else	  
          printf("\ntangentsPerFace: Zero tangent - [%g,%g,%g]!!! Exiting...\n",tangentsF[i][0],tangentsF[i][1],tangentsF[i][2]);
          exit(-1);
#endif	  
        }

      if(isnan(tangentsF[i][0]) || isnan(tangentsF[i][0]) || isnan(tangentsF[i][0]))
        {
#ifdef HACK
	  tangentsF[i][0] = 1.0;
	  tangentsF[i][1] = 0.0;
	  tangentsF[i][2] = 0.0;
#else	  
          printf("\ntangentsPerFace: NaN in tangent at face: %d - [%g,%g,%g]. Exiting...\n",i,tangentsF[i][0],tangentsF[i][1],tangentsF[i][2]);
          printf("                      UV0   - [%g,%g]\n",UV[0][0],UV[0][1]);
          printf("                      UV1   - [%g,%g]\n",UV[1][0],UV[1][1]);
          printf("                      UV2   - [%g,%g]\n\n",UV[2][0],UV[2][1]);
          printf("                      V0   - [%g,%g,%g]\n",v[0][0],v[0][1],v[0][2]);
          printf("                      V1   - [%g,%g,%g]\n",v[1][0],v[1][1],v[1][2]);
          printf("                      V2   - [%g,%g,%g]\n",v[2][0],v[2][1],v[2][2]);
          exit(-1);
#endif
        }
    }// for faces
  printf("Computing tangents for individual faces...done\n");fflush(stdout);
  freemem2(UV, 0,2, 0,2); 
  freemem2(v,  0,2, 0,2); 
  freemem2(n,  0,2, 0,2); 
  return;
}//--- tangentsPerFace ---------------------------------------------------------------

void TObjectData::tangentsPerVertex()
{
  printf("Computing tangents for individual vertices...      \r");fflush(stdout);
  if(tangentsV==NULL)
    tangentsV = allocation2(1,dataSize,0,2); 
  else
    return;

  strcpy(fileName,OBJfileName);
  fileName[strlen(OBJfileName)-4] = '\0';
  strcat(fileName,".tan");

  bool status = loadTangentsPerVertex(fileName);
  printf("Loading status: '%s' (%d)\n",fileName, status);
  if(status == false)
    {
      if(tangentsF==NULL)
        this->tangentsPerFace();
      // computing tangents for individual vertices
      int trianglesPerVertex;
      float v[3][3];
      for(int i=1;i<i_v;i++)
        {
          for(int j=0;j<=2;j++)
            tangentsV[i][j] = 0.f;
          trianglesPerVertex = 0;
          for(int k=1;k<i_f;k++)
            {
              
              for(int j=0;j<=2;j++)
                {
                  v[0][j] = vertices[faces[k][0]][j];
                  v[1][j] = vertices[faces[k][1]][j];
                  v[2][j] = vertices[faces[k][2]][j];
                }
              if( ((v[0][0] == vertices[i][0])&&(v[0][1] == vertices[i][1])&&(v[0][2] == vertices[i][2]))||
                  ((v[1][0] == vertices[i][0])&&(v[1][1] == vertices[i][1])&&(v[1][2] == vertices[i][2]))||
                  ((v[2][0] == vertices[i][0])&&(v[2][1] == vertices[i][1])&&(v[2][2] == vertices[i][2])) )
                
                //                      if((i == faces[k][0])||(i == faces[k][1])||(i == faces[k][2]))
                {
                  for(int j=0;j<=2;j++)
                    tangentsV[i][j] += tangentsF[k][j];
                  trianglesPerVertex++;
                }
            }// for faces
          if(trianglesPerVertex >= 1)
            for(int j=0;j<=2;j++)
              tangentsV[i][j] /= trianglesPerVertex;
          else
            for(int j=0;j<=2;j++)
              tangentsV[i][j] = 0.1f;
          
          if(isnan(tangentsV[i][0]) || isnan(tangentsV[i][0]) || isnan(tangentsV[i][0])) 
            {
              printf("\ntangentsPerVertex: NaN in tangent at face: %d - [%g,%g,%g] %d. Exiting...\n",i,tangentsV[i][0],tangentsV[i][1],tangentsV[i][2], trianglesPerVertex);
              exit(-1);
            }
          
          if(vec_len(tangentsV[i])<=0.f) 
            {
#ifdef HACK
	  tangentsV[i][0] = 1.0;
	  tangentsV[i][1] = 0.0;
	  tangentsV[i][2] = 0.0;
#else	  	      
              printf("\ntangentsPerVertex: Zero tangent at face: %d - [%g,%g,%g]!!! Exiting...\n",i,tangentsV[i][0],tangentsV[i][1],tangentsV[i][2]);	      
              exit(-1);
#endif	      
            }
          vec_norm(tangentsV[i]);
        }// for vertices
      printf("Computing tangents for individual vertices...done  \n");fflush(stdout);
      
      // saving computed tangents (if not done already)
      printf("Saving tangents per vertex                   \r");
      saveTangentsPerVertex(fileName);
    }

  return;
}//--- tangentsPerVertex ----------------------------------------------------------------

void TObjectData::saveTangentsPerVertex(char *fileName)
{
  FILE *fp;
  
  printf("\nSaving: %s                                \n",fileName);

  if((fp=fopen(fileName,"w"))==NULL)
    {
      printf(" Error when creating file: '%s' !!!\n",fileName);
      exit(-1);
    }

  fprintf(fp,"%d\n", (int)i_v-1);
  for(int i=1;i<i_v;i++)
    fprintf(fp,"%f %f %f\n",tangentsV[i][0],tangentsV[i][1],tangentsV[i][2]);
  
  fclose(fp);    
  printf(" Computed tangents per vertex were saved to file: '%s'\n",fileName);

  return;
}//--- saveTangentsPerVertex ------------------------------------------------------------

bool TObjectData::loadTangentsPerVertex(char *fileName)
{
  FILE *fp;
  
  printf("Loading: %s                                \n",fileName);

  if((fp=fopen(fileName,"r"))==NULL)
    {
      printf(" Cannot open file: '%s'. Computing tangents instead...\n",fileName);
      return false;
    }
  int NoOfTangents;
  int tt = fscanf(fp,"%d\n", &NoOfTangents);
  if(NoOfTangents > i_v-1) { 
    printf("Wrong number of tangents %d (assumed %d) !!! Exiting...\n", (int)NoOfTangents, (int)i_v);
    exit(-1);
  }
  for(int i=1;i<=NoOfTangents;i++) {
    tt = fscanf(fp,"%f %f %f\n",&tangentsV[i][0],&tangentsV[i][1],&tangentsV[i][2]);
  }
  
  fclose(fp);    
  
  return true;
}//--- loadTangentsPerVertex ------------------------------------------------------------

void TObjectData::binormalsPerFace()
{
  printf("Computing binormals for individual faces...      \r");fflush(stdout);
  if(binormalsF==NULL)
    binormalsF = allocation2(1,dataSize,0,2); 
  else
    return;
  if(normalsF==NULL)
    this->normalsPerFace();
  if(tangentsF==NULL)
    this->tangentsPerFace();
  // computing binormals
  for(int i=1;i<i_f;i++)
    {
      for(int k=0;k<=2;k++)
        {
          vec_crossprod(normalsF[i],tangentsF[i], binormalsF[i]);

          if(vec_len(binormalsF[i])<=0.f) 
            {
              printf("binormalsPerFace: Zero binormal - [%g,%g,%g]!!! Exiting...\n",binormalsF[i][0],binormalsF[i][1],binormalsF[i][2]);
              exit(-1);
            }
          vec_norm(binormalsF[i]);
        }
    }
  printf("Computing binormals for individual faces...done   \n");fflush(stdout);
  return;
}//--- binormalsPerFace ------------------------------------------------------------------

void TObjectData::binormalsPerVertex()
{
  if(i_vn==1)
    {
      printf("OBJloader: !!! Normals per vertex are not available in the model !!! Exiting...\n");
      exit(-1);
    }
  printf("Computing binormals for individual vertices...      \r");fflush(stdout);
  if(binormalsV==NULL)
    binormalsV = allocation2(1,dataSize,0,2); 
  else
    return;
  if(tangentsV==NULL)
    this->tangentsPerVertex();
  // computing binormals
  for(int i=1;i<i_f;i++)
    {
      for(int k=0;k<=2;k++)
        {
          if(normalsV==NULL)
            vec_crossprod(normals[faces[i][6+k]],tangentsV[faces[i][k]], binormalsV[faces[i][k]]);
          else
            vec_crossprod(normalsV[faces[i][k]],tangentsV[faces[i][k]], binormalsV[faces[i][k]]);

          if(vec_len(binormalsV[faces[i][k]])<=0.f) 
            {
              printf("binormalsPerVertex: Zero binormal - [%g,%g,%g]!!! Exiting...\n",binormalsV[faces[i][k]][0],binormalsV[faces[i][k]][1],binormalsV[faces[i][k]][2]);
              printf("%f %f %f\n",tangentsV[faces[i][k]][0],tangentsV[faces[i][k]][1],tangentsV[faces[i][k]][2]);
              printf("%f %f %f\n",normals[faces[i][k+6]][0],normals[faces[i][k+6]][1],normals[faces[i][k+6]][2]);
              exit(-1);
            }
          vec_norm(binormalsV[faces[i][k]]);
        }
    }
  printf("Computing binormals for individual vertices...done   \n");fflush(stdout);
  return;
}//--- binormalsPerVertex --------------------------------------------------------------

float TObjectData::getV(int face, int vertex, int xyz)
{
  return vertices[faces[face][vertex]][xyz];
}//--- getV ----------------------------------------------------------------------------

int TObjectData::getVindex(int face, int vertex)
{
  return faces[face][vertex];
}//--- getVindex -----------------------------------------------------------------------

float TObjectData::getN(int face, int vertex, int xyz)
{
  if(i_vn>1)
    return normals[faces[face][6+vertex]][xyz];
  else
    {
      printf("OBJloader: !!! Normals per vertex are not available in the model !!! Exiting...\n");
      exit(-1);
    }
}//--- getN ----------------------------------------------------------------------------

float TObjectData::getNv(int face, int vertex, int xyz)
{
  if(normalsV!=NULL)
    return normalsV[faces[face][6+vertex]][xyz];
  else
    {
      printf("OBJloader: !!! Normals per vertex are not computed !!! Exiting...\n");
      exit(-1);
    }
}//--- getNv ----------------------------------------------------------------------------

float TObjectData::getNf(int face, int xyz)
{
  if(normalsF!=NULL)
    return normalsF[face][xyz];
  else
    {
      printf("OBJloader: !!! Normals per face are not computed !!! Exiting...\n");
      exit(-1);
    }
}//--- getNf ---------------------------------------------------------------------------

float TObjectData::getT(int face, int vertex, int xyz)
{
  if(tangentsV!=NULL)
    return tangentsV[faces[face][vertex]][xyz];
  else
    {
      printf("OBJloader: !!! Tangents per vertex are not computed !!! Exiting...\n");
      exit(-1);
    }

}//--- getT ----------------------------------------------------------------------------

float TObjectData::getTf(int face, int xyz)
{
  if(tangentsF!=NULL)
  return tangentsF[face][xyz];
  else
    {
      printf("OBJloader: !!! Tangents per face are not computed !!! Exiting...\n");
      exit(-1);
    }
}//--- getTf ---------------------------------------------------------------------------

float TObjectData::getB(int face, int vertex, int xyz)
{
  if(binormalsV!=NULL)
  return binormalsV[faces[face][vertex]][xyz];
  else
    {
      printf("OBJloader: !!! Binormals per vertex are not computed !!! Exiting...\n");
      exit(-1);
    }
}//--- getB ----------------------------------------------------------------------------

float TObjectData::getBf(int face, int xyz)
{
  if(binormalsF!=NULL)
  return binormalsF[face][xyz];
  else
    {
      printf("OBJloader: !!! Binormals per face are not computed !!! Exiting...\n");
      exit(-1);
    }
}//--- getB ----------------------------------------------------------------------------

float TObjectData::getUV(int face, int vertex, int uv)
{
  if(i_vt>1)
    return UVcoord[faces[face][3+vertex]][uv];
  else
    {
      printf("OBJloader: !!! UV coordinates are not available in the model !!! Exiting...\n");
      exit(-1);
    }
}//--- getUV ----------------------------------------------------------------------------


//###############################################################
// 3D Vector operations
//###############################################################

void TObjectData::vec_crossprod(float *x, float *y, float *res)
{
    res[0] = x[1]*y[2] - x[2]*y[1];
    res[1] = x[2]*y[0] - x[0]*y[2];
    res[2] = x[0]*y[1] - x[1]*y[0];
}//--- vec_crossprod ---------------------------------------------

float TObjectData::vec_len(float *x)
{
    return sqrtf(x[0]*x[0] + x[1]*x[1] + x[2]*x[2]);
}//--- vec_len ---------------------------------------------------

void TObjectData::vec_diff(float *x, float *y, float *res)
{
    res[0] = x[0] - y[0];
    res[1] = x[1] - y[1];
    res[2] = x[2] - y[2];
}//--- vec_diff ---------------------------------------------------

int TObjectData::vec_norm(float *x)
{
    float aux;
    aux = vec_len(x);
    if(aux == 0.f)
      {		
        x[0] = 0.f;
        x[1] = 0.f;
        x[2] = 0.f;
        printf("vec_norm: zero length vector");    
        return 0;
      }
    x[0] = x[0]/aux;
    x[1] = x[1]/aux;
    x[2] = x[2]/aux;
    return 1;
}//--- vec_norm ---------------------------------------------------

void TObjectData::normal(float *v1, float *v2, float *v3, float *n)
{
    float *p, *q;
    p = allocation1(0,2);
    q = allocation1(0,2);
    
    vec_diff(v1,v2,p);
    vec_diff(v3,v2,q);
    vec_crossprod(q,p,n);
    vec_norm(n);

    freemem1(p,0,2);
    freemem1(q,0,2);
}//--- normal -----------------------------------------------------
