// ================================================================
// $Id: $
//
// readpng.cpp
//
// This implementss PNG-reading function
//
// Initial coding by Jiri Filip, 2004.

#include <cstdlib>
#include <cstdio>

#include "_libpng/png.h"

#define PI 3.14159265358979323846

/************************************************************************
 from/for Pres.cpp -> image IO definitions: -----------------------------
 ************************************************************************/
#define IMGIO_FILE_NOT_OPENED -1
#define IMGIO_NOT_A_PNG -2
#define IMGIO_ALLOCATION_PROBLEM -3
#define IMGIO_FILE_READ_PROBLEM -4
#define IMGIO_FILE_WRITE_PROBLEM -5
#define IMGIO_FILE_CONTENTS_PROBLEM -6
#define IMGIO_WRONG_PARAMETER -7

#define PNG_BYTES_TO_CHECK 4

int
readPNGheader(const char *filename, int *planes, int *nr, int *nc)
{
   /*! the procedure only reads basic image information from the 'filename' PNG file */

   char buf[PNG_BYTES_TO_CHECK];
   png_uint_32 width, height;
   int bit_depth, color_type, interlace_type;
   png_structp png_ptr;
   png_infop info_ptr /*, endinfo_ptr*/;
   FILE *fp = fopen(filename, "rb");
   if (!fp)
   {
       /*printf("File cannot be opened.");*/
       return IMGIO_FILE_NOT_OPENED;
   }
   if (fread(buf, 1, PNG_BYTES_TO_CHECK, fp) != PNG_BYTES_TO_CHECK) return IMGIO_FILE_NOT_OPENED;
   if (png_sig_cmp((png_bytep)buf, (png_size_t)0, PNG_BYTES_TO_CHECK))
   {
       /*printf("Not a PNG file.");*/
       return IMGIO_NOT_A_PNG;
   }

   /* Create and initialize the png_struct with the desired error handler
    * functions.  If you want to use the default stderr and longjump method,
    * you can supply NULL for the last three parameters.  We also supply the
    * the compiler header file version, so that we know if the application
    * was compiled with a compatible version of the library.  REQUIRED
    */
   png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, (png_voidp)NULL, NULL, NULL);
   if (!png_ptr)
   {
       /*printf("Error creating PNG structures.");*/
       return IMGIO_ALLOCATION_PROBLEM;
   }

   /* Allocate/initialize the memory for image information.  REQUIRED. */
   info_ptr = png_create_info_struct(png_ptr);
   if (!info_ptr)
   {
       png_destroy_read_struct(&png_ptr,
          (png_infopp)NULL, (png_infopp)NULL);
       /*printf("Error creating PNG structures.");*/
       return IMGIO_ALLOCATION_PROBLEM;
   }
    
   /* Set error handling if you are using the setjmp/longjmp method (this is
    * the normal method of doing things with libpng).  REQUIRED unless you
    * set up your own error handlers in the png_create_read_struct() earlier.
    */
   if (setjmp(png_jmpbuf(png_ptr)))
   {
      /* Free all of the memory associated with the png_ptr and info_ptr */
      png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
      fclose(fp);
      /* If we get here, we had a problem reading the file */
      return IMGIO_FILE_READ_PROBLEM;
   }
   
   /* Set up the input control if you are using standard C streams */
   png_init_io(png_ptr, fp);

   png_set_sig_bytes(png_ptr,PNG_BYTES_TO_CHECK);
   
   /* The call to png_read_info() gives us all of the information from the
    * PNG file before the first IDAT (image data chunk).  REQUIRED
    */
   png_read_info(png_ptr, info_ptr);
   png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type, &interlace_type, int_p_NULL, int_p_NULL);
   *nc=width;
   *nr=height;
       
   if((color_type == PNG_COLOR_TYPE_GRAY) || (color_type == PNG_COLOR_TYPE_GRAY_ALPHA)) *planes=1;
   else
   if((color_type == PNG_COLOR_TYPE_RGB) || (color_type == PNG_COLOR_TYPE_RGB_ALPHA) || (color_type == PNG_COLOR_TYPE_PALETTE)) *planes=3;
   else return -1; /* not a usable color_type */

   /* Free all of the memory associated with the png_ptr and info_ptr */
   png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
   fclose(fp);
   return 0;
} /* readPNGheader ------------------------------------------------------ */

int readPNGfile (const char *filename,unsigned char *arr,int planes, int nr, int nc, int oneplane)
{
   /*! \param 'planes' should be 1 for gray-scale or 3 for RGB */
   /*! \note this filter will automatically convert between rgb and grayscale
      images in case of need, depending on the user-specified 'planes' value */
   
   /*! \warning Gamma information is currently ignored !
       \warning Alpha information is currently ignored ! (black background substituted) */

   int imgInd;

   char buf[PNG_BYTES_TO_CHECK];
   png_uint_32 width, height;
   int bit_depth, color_type, interlace_type;
   int number_passes; /* for interlaced images */
   int x,y,xi;
   png_structp png_ptr;
   png_infop info_ptr /*, endinfo_ptr*/;
   png_bytep row;
   png_bytepp rows;
   //   float **arr2;   
   FILE *fp = fopen(filename, "rb");

   /*
   if(planes==1) // grayscale image 
   {
      if(oneplane<0) arr2=(float**)arr; else arr2 = arr[oneplane];
   }
   */

   if (!fp)
   {
       /*printf("File cannot be opened.");*/
       return IMGIO_FILE_NOT_OPENED;
   }
   if (fread(buf, 1, PNG_BYTES_TO_CHECK, fp) != PNG_BYTES_TO_CHECK) return IMGIO_FILE_NOT_OPENED;
   if (png_sig_cmp((png_bytep)buf, (png_size_t)0, PNG_BYTES_TO_CHECK))
   {
       /*printf("Not a PNG file.");*/
       return IMGIO_NOT_A_PNG;
   }

   /* Create and initialize the png_struct with the desired error handler
    * functions.  If you want to use the default stderr and longjump method,
    * you can supply NULL for the last three parameters.  We also supply the
    * the compiler header file version, so that we know if the application
    * was compiled with a compatible version of the library.  REQUIRED
    */
   png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, (png_voidp)NULL, NULL, NULL);
   if (!png_ptr)
   {
       /*printf("Error creating PNG structures.");*/
       return IMGIO_ALLOCATION_PROBLEM;
   }

   /* Allocate/initialize the memory for image information.  REQUIRED. */
   info_ptr = png_create_info_struct(png_ptr);
   if (!info_ptr)
   {
       png_destroy_read_struct(&png_ptr,
          (png_infopp)NULL, (png_infopp)NULL);
       /*printf("Error creating PNG structures.");*/
       return IMGIO_ALLOCATION_PROBLEM;
   }
    
   /* Set error handling if you are using the setjmp/longjmp method (this is
    * the normal method of doing things with libpng).  REQUIRED unless you
    * set up your own error handlers in the png_create_read_struct() earlier.
    */
   if (setjmp(png_jmpbuf(png_ptr)))
   {
      /* Free all of the memory associated with the png_ptr and info_ptr */
      png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
      fclose(fp);
      /* If we get here, we had a problem reading the file */
      return IMGIO_FILE_READ_PROBLEM;
   }
   
   /* Set up the input control if you are using standard C streams */
   png_init_io(png_ptr, fp);

   png_set_sig_bytes(png_ptr,PNG_BYTES_TO_CHECK);
   
   /* The call to png_read_info() gives us all of the information from the
    * PNG file before the first IDAT (image data chunk).  REQUIRED
    */
   png_read_info(png_ptr, info_ptr);
   png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
       &interlace_type, int_p_NULL, int_p_NULL);
 
   /* tell libpng to strip 16 bit/color files down to 8 bits/color */
   png_set_strip_16(png_ptr);

   /* this would simply remove the alpha channel, if present.
      Currently the alpha-data removal is postponed to the time of raw
      data transformation to ***arr. This needs more memory (by 33%), but
      allows further alpha processing if needed. */
   /*if (color_type & PNG_COLOR_MASK_ALPHA) png_set_strip_alpha(png_ptr);*/

   /* Extract multiple pixels with bit depths of 1, 2, and 4 from a single
    * byte into separate bytes (useful for paletted and grayscale images).
    */
   png_set_packing(png_ptr);

   /* in case the png file is grayscale and the user requests a rgb image,
      transform automatically the image data */
   if ((planes==3)&&(color_type == PNG_COLOR_TYPE_GRAY ||
       color_type == PNG_COLOR_TYPE_GRAY_ALPHA))
         png_set_gray_to_rgb(png_ptr);

   /* in case the png file is rgb and the user requests a grayscale image,
      transform automatically the image data */
   if ((planes==1)&&(color_type == PNG_COLOR_TYPE_RGB ||
       color_type == PNG_COLOR_TYPE_RGB_ALPHA))
         png_set_rgb_to_gray_fixed(png_ptr, 1 /*error_action=silent*/, -1 /*red_weight=default*/, -1 /*green_weight=default*/);

   if((width!=nc)||(height!=nr))
   {
      /* Free all of the memory associated with the png_ptr and info_ptr */
      png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
      fclose(fp);
      /* some file properties are different from the expected */
      return IMGIO_FILE_CONTENTS_PROBLEM;
   }

   /* Expand paletted colors into true RGB triplets */
   if ((planes==3)&&(color_type == PNG_COLOR_TYPE_PALETTE)) // \warning (planes==3) added by PS, otherwise gray images become corrupted during input
      png_set_palette_to_rgb(png_ptr);

   /* Expand grayscale images to the full 8 bits from 1, 2, or 4 bits/pixel */
   if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
      png_set_gray_1_2_4_to_8(png_ptr);

   /* Turn on interlace handling.  REQUIRED if you are not using
    * png_read_image().  To see how to handle interlacing passes,
    * see the png_read_row() method below:
    */
   number_passes = png_set_interlace_handling(png_ptr);

   /* Optional call to gamma correct and add the background to the palette
    * and update info structure.  REQUIRED if you are expecting libpng to
    * update the palette for you (ie you selected such a transform above).
    */
   png_read_update_info(png_ptr, info_ptr);

   if(number_passes==1)
   { /* one row at a time - less memory needed */
      /* Allocate the memory to hold one row at a time using the fields of info_ptr. */
      row = (png_byte *)png_malloc(png_ptr, png_get_rowbytes(png_ptr,info_ptr));
      if (row==NULL)
      {
         /* Free all of the memory associated with the png_ptr and info_ptr */
         png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
         fclose(fp);
         return IMGIO_ALLOCATION_PROBLEM;
      }

      imgInd = 0;

      for (y = 0; y < height; y++)
      {
         png_read_rows(png_ptr, &row, png_bytepp_NULL, 1);
         
         /* copying to arr */
         if(planes==1) /* one plane */
         {
      		for(x=0;x<width;x++) 
      		{
              //      			arr2[y][x]=row[xi++];
              arr[imgInd++] = row[x];
              arr[imgInd++] = row[x];
              arr[imgInd++] = row[x];
      			/* currently the (possible) alpha channel is ignored, i.e. the
      			   resulting image data looks like combined with black backgroung */
              if(color_type & PNG_COLOR_MASK_ALPHA) {
                arr[imgInd++]=row[x];
              }
              else
                arr[imgInd++]=0xff;
                //      			if(color_type & PNG_COLOR_MASK_ALPHA) xi++;
      		}
         }
         else /* 3 planes */
         {
      		xi=0;
            for(x=0;x<width;x++)
              {
      			arr[imgInd++]=row[xi++];
      			arr[imgInd++]=row[xi++];
      			arr[imgInd++]=row[xi++];
      			/* currently the (possible) alpha channel is ignored, i.e. the
      			   resulting image data looks like combined with black backgroung */
                if(color_type & PNG_COLOR_MASK_ALPHA) {
                  arr[imgInd++]=row[xi++];
                }
                else
                  arr[imgInd++]=0xff;
              }
         }
      }
      png_free(png_ptr,row);
   }
   else /* interlaced images must be read all at once */
   {
      rows = (png_byte **)png_malloc(png_ptr, height*sizeof(png_bytep));
      if (rows==NULL)
      {
         /* Free all of the memory associated with the png_ptr and info_ptr */
         png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
         fclose(fp);
         return IMGIO_ALLOCATION_PROBLEM;
      }
      for (y = 0; y < height; y++)
      {
         rows[y] = (png_byte *)png_malloc(png_ptr, png_get_rowbytes(png_ptr,info_ptr));
         if (rows[y]==NULL)
         {
            /* Free all of the memory associated with the png_ptr and info_ptr */
            png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
            fclose(fp);
            return IMGIO_ALLOCATION_PROBLEM;
         }
      }
         
      png_read_image(png_ptr, rows);

      
      /* copying to arr */
      if(planes==1) /* one plane */
      {
      	for(y=0;y<height;y++)
      	{
      		for(x=0;x<width;x++) 
      		{
              //      			arr2[y][x]=rows[y][xi++];
              arr[imgInd++] = row[x];
              arr[imgInd++] = row[x];
              arr[imgInd++] = row[x];
      			/* currently the (possible) alpha channel is ignored, i.e. the
      			   resulting image data looks like combined with black backgroung */
              //      			if(color_type & PNG_COLOR_MASK_ALPHA) xi++;
              if(color_type & PNG_COLOR_MASK_ALPHA) {
                arr[imgInd++]=row[x];
              }
              else
                arr[imgInd++]=0xff;
      		}
      	}
      }
      else /* 3 planes */
      {
      	for(y=0;y<height;y++)
      	{
          xi=0;
          for(x=0;x<width;x++)
            {
              arr[imgInd++]=row[xi++];
              arr[imgInd++]=row[xi++];
              arr[imgInd++]=row[xi++];

              /* currently the (possible) alpha channel is ignored, i.e. the
                 resulting image data looks like combined with black backgroung */
              if(color_type & PNG_COLOR_MASK_ALPHA) {
                arr[imgInd++]=row[xi++];
              }
              else
                arr[imgInd++]=0xff;

              }
      	}
      }
      
      for (y = 0; y < height; y++)
         png_free(png_ptr,rows[y]);
      png_free(png_ptr,rows);
   }

   /* read rest of file, and get additional chunks in info_ptr - REQUIRED */
   png_read_end(png_ptr, info_ptr);

   /* At this point you have read the entire image */
   /* clean up after the read, and free any memory allocated - REQUIRED */
   png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);

   fclose(fp);  
   return 1;
} /* readPNGfile ------------------------------------------------------ */


int
readPNGfile3D(const char *filename,float ***arr,int planes, int nr, int nc, int oneplane)
{
   /*! \param 'planes' should be 1 for gray-scale or 3 for RGB */
   /*! \note this filter will automatically convert between rgb and grayscale
      images in case of need, depending on the user-specified 'planes' value */

   /*! \warning Gamma information is currently ignored !
       \warning Alpha information is currently ignored ! (black background substituted) */

   char buf[PNG_BYTES_TO_CHECK];
   png_uint_32 width, height;
   int bit_depth, color_type, interlace_type;
   int number_passes; /* for interlaced images */
   int x,y,xi;
   png_structp png_ptr;
   png_infop info_ptr /*, endinfo_ptr*/;
   png_bytep row;
   png_bytepp rows;
   float **arr2=NULL;
   unsigned short byte1,byte2;

   //union {unsigned short ipix; char cpix[2];} bitconv;
   FILE *fp = fopen(filename, "rb");

   if(planes==1) /* grayscale image */
   {
      if(oneplane<0) arr2=(float**)arr; else arr2 = arr[oneplane];
   }

   if (!fp)
   {
       /*printf("File cannot be opened.");*/
       return IMGIO_FILE_NOT_OPENED;
   }
   if (fread(buf, 1, PNG_BYTES_TO_CHECK, fp) != PNG_BYTES_TO_CHECK) return IMGIO_FILE_NOT_OPENED;
   if (png_sig_cmp((png_bytep)buf, (png_size_t)0, PNG_BYTES_TO_CHECK))
   {
       /*printf("Not a PNG file.");*/
       return IMGIO_NOT_A_PNG;
   }

   /* Create and initialize the png_struct with the desired error handler
    * functions.  If you want to use the default stderr and longjump method,
    * you can supply NULL for the last three parameters.  We also supply the
    * the compiler header file version, so that we know if the application
    * was compiled with a compatible version of the library.  REQUIRED
    */
   png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, (png_voidp)NULL, NULL, NULL);
   if (!png_ptr)
   {
       /*printf("Error creating PNG structures.");*/
       return IMGIO_ALLOCATION_PROBLEM;
   }

   /* Allocate/initialize the memory for image information.  REQUIRED. */
   info_ptr = png_create_info_struct(png_ptr);
   if (!info_ptr)
   {
       png_destroy_read_struct(&png_ptr,
          (png_infopp)NULL, (png_infopp)NULL);
       /*printf("Error creating PNG structures.");*/
       return IMGIO_ALLOCATION_PROBLEM;
   }

   /* Set error handling if you are using the setjmp/longjmp method (this is
    * the normal method of doing things with libpng).  REQUIRED unless you
    * set up your own error handlers in the png_create_read_struct() earlier.
    */
   if (setjmp(png_jmpbuf(png_ptr)))
   {
      /* Free all of the memory associated with the png_ptr and info_ptr */
      png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
      fclose(fp);
      /* If we get here, we had a problem reading the file */
      return IMGIO_FILE_READ_PROBLEM;
   }

   /* Set up the input control if you are using standard C streams */
   png_init_io(png_ptr, fp);

   png_set_sig_bytes(png_ptr,PNG_BYTES_TO_CHECK);

   /* The call to png_read_info() gives us all of the information from the
    * PNG file before the first IDAT (image data chunk).  REQUIRED
    */
   png_read_info(png_ptr, info_ptr);
   png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
       &interlace_type, int_p_NULL, int_p_NULL);
   
   /* tell libpng to strip 16 bit/color files down to 8 bits/color */
   /*png_set_strip_16(png_ptr);*/

   /* this would simply remove the alpha channel, if present.
      Currently the alpha-data removal is postponed to the time of raw
      data transformation to ***arr. This needs more memory (by 33%), but
      allows further alpha processing if needed. */
   /*if (color_type & PNG_COLOR_MASK_ALPHA) png_set_strip_alpha(png_ptr);*/

   /* Extract multiple pixels with bit depths of 1, 2, and 4 from a single
    * byte into separate bytes (useful for paletted and grayscale images).
    */
   png_set_packing(png_ptr);

   /* in case the png file is grayscale and the user requests a rgb image,
      transform automatically the image data */
   if ((planes==3)&&(color_type == PNG_COLOR_TYPE_GRAY ||
       color_type == PNG_COLOR_TYPE_GRAY_ALPHA))
         png_set_gray_to_rgb(png_ptr);

   /* in case the png file is rgb and the user requests a grayscale image,
      transform automatically the image data */
   if ((planes==1)&&(color_type == PNG_COLOR_TYPE_RGB ||
       color_type == PNG_COLOR_TYPE_RGB_ALPHA))
         png_set_rgb_to_gray_fixed(png_ptr, 1 /*error_action=silent*/, -1 /*red_weight=default*/, -1 /*green_weight=default*/);

   if((width!=nc)||(height!=nr))
   {
      /* Free all of the memory associated with the png_ptr and info_ptr */
      png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
      fclose(fp);
      /* some file properties are different from the expected */
      return IMGIO_FILE_CONTENTS_PROBLEM;
   }

   /* Expand paletted colors into true RGB triplets */
   if ((planes==3)&&(color_type == PNG_COLOR_TYPE_PALETTE)) /* \warning (planes==3) added by PS, otherwise gray images become corrupted during input*/
      png_set_palette_to_rgb(png_ptr);

   /* Expand grayscale images to the full 8 bits from 1, 2, or 4 bits/pixel */
   if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
      png_set_gray_1_2_4_to_8(png_ptr);

   /* Turn on interlace handling.  REQUIRED if you are not using
    * png_read_image().  To see how to handle interlacing passes,
    * see the png_read_row() method below:
    */
   number_passes = png_set_interlace_handling(png_ptr);

   /* Optional call to gamma correct and add the background to the palette
    * and update info structure.  REQUIRED if you are expecting libpng to
    * update the palette for you (ie you selected such a transform above).
    */
   png_read_update_info(png_ptr, info_ptr);

   if(number_passes==1)
   { /* one row at a time - less memory needed */
      /* Allocate the memory to hold one row at a time using the fields of info_ptr. */
      row = (png_byte *)png_malloc(png_ptr, png_get_rowbytes(png_ptr,info_ptr));
      if (row==NULL)
      {
         /* Free all of the memory associated with the png_ptr and info_ptr */
         png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
         fclose(fp);
         return IMGIO_ALLOCATION_PROBLEM;
      }

 	for (y = 0; y < height; y++)
      {
         png_read_rows(png_ptr, &row, png_bytepp_NULL, 1);

         /* copying to arr */
         if(planes==1) /* one plane */
         {
         	xi=0;
      		for(x=0;x<width;x++) if(bit_depth==16)
      		{
				byte1 = row[xi++];
				byte2 = row[xi++];
				byte2 <<= 8;
				arr2[y][x]=(float)(byte1+byte2);

      			/* currently the (possible) alpha channel is ignored, i.e. the
      			   resulting image data looks like combined with black backgroung */
      			if(color_type & PNG_COLOR_MASK_ALPHA) xi+=2;
      		} else {
      			arr2[y][x]=(float)row[xi++];
      			/* currently the (possible) alpha channel is ignored, i.e. the
      			   resulting image data looks like combined with black backgroung */
      			if(color_type & PNG_COLOR_MASK_ALPHA) xi++;
      		}
         }
         else /* 3 planes */
         {
      		xi=0;
      		for(x=0;x<width;x++) if(bit_depth==16)
      		{
				byte1 = row[xi++];
				byte2 = row[xi++];
				byte2 <<= 8;
				arr[0][y][x]=(float)(byte1+byte2);
				byte1 = row[xi++];
				byte2 = row[xi++];
				byte2 <<= 8;
				arr[1][y][x]=(float)(byte1+byte2);
				byte1 = row[xi++];
				byte2 = row[xi++];
				byte2 <<= 8;
				arr[2][y][x]=(float)(byte1+byte2);

/*
      			bitconv.cpix[0]=row[xi++];
      			bitconv.cpix[1]=row[xi++];
      			arr[0][y][x]=(float)bitconv.ipix;
      			bitconv.cpix[0]=row[xi++];
      			bitconv.cpix[1]=row[xi++];
      			arr[1][y][x]=(float)bitconv.ipix;
      			bitconv.cpix[0]=row[xi++];
      			bitconv.cpix[1]=row[xi++];
      			arr[2][y][x]=(float)bitconv.ipix;
				*/
      			/* currently the (possible) alpha channel is ignored, i.e. the
      			   resulting image data looks like combined with black backgroung */
      			if(color_type & PNG_COLOR_MASK_ALPHA) xi+=2;
      		} else {
      			arr[0][y][x]=(float)row[xi++];
      			arr[1][y][x]=(float)row[xi++];
      			arr[2][y][x]=(float)row[xi++];
      			/* currently the (possible) alpha channel is ignored, i.e. the
      			   resulting image data looks like combined with black backgroung */
      			if(color_type & PNG_COLOR_MASK_ALPHA) xi++;
      		}
         }
      }
      png_free(png_ptr,row);
   }
   else /* interlaced images must be read all at once */
   {
      rows = (png_byte **)png_malloc(png_ptr, height*sizeof(png_bytep));
      if (rows==NULL)
      {
         /* Free all of the memory associated with the png_ptr and info_ptr */
         png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
         fclose(fp);
         return IMGIO_ALLOCATION_PROBLEM;
      }
      for (y = 0; y < height; y++)
      {
         rows[y] = (png_byte *)png_malloc(png_ptr, png_get_rowbytes(png_ptr,info_ptr));
         if (rows[y]==NULL)
         {
            /* Free all of the memory associated with the png_ptr and info_ptr */
            png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
            fclose(fp);
            return IMGIO_ALLOCATION_PROBLEM;
         }
      }

      png_read_image(png_ptr, rows);

      /* copying to arr */
      if(planes==1) /* one plane */
      {
      	for(y=0;y<height;y++)
      	{
      		xi=0;
      		for(x=0;x<width;x++) if(bit_depth==16)
      		{
      			byte1 = rows[y][xi++];
				byte2 = rows[y][xi++];
				byte2 <<= 8;
				arr2[y][x]=(float)(byte1+byte2);
				/*
				bitconv.cpix[0]=rows[y][xi++];
      			bitconv.cpix[1]=rows[y][xi++];
      			arr2[y][x]=(float)bitconv.ipix;
				*/
      			/* currently the (possible) alpha channel is ignored, i.e. the
      			   resulting image data looks like combined with black backgroung */
      			if(color_type & PNG_COLOR_MASK_ALPHA) xi+=2;
      		} else {
      			arr2[y][x]=(float)rows[y][xi++];
      			/* currently the (possible) alpha channel is ignored, i.e. the
      			   resulting image data looks like combined with black backgroung */
      			if(color_type & PNG_COLOR_MASK_ALPHA) xi++;
      		}
      	}
      }
      else /* 3 planes */
      {
      	for(y=0;y<height;y++)
      	{
      		xi=0;
      		for(x=0;x<width;x++) if(bit_depth==16)
      		{
				byte1 = rows[y][xi++];
				byte2 = rows[y][xi++];
				byte2 <<= 8;
				arr[0][y][x]=(float)(byte1+byte2);
     			byte1 = rows[y][xi++];
				byte2 = rows[y][xi++];
				byte2 <<= 8;
				arr[1][y][x]=(float)(byte1+byte2);
     			byte1 = rows[y][xi++];
				byte2 = rows[y][xi++];
				byte2 <<= 8;
				arr[2][y][x]=(float)(byte1+byte2);
     			/*
				bitconv.cpix[0]=rows[y][xi++];
      			bitconv.cpix[1]=rows[y][xi++];
      			arr[0][y][x]=(float)bitconv.ipix;
      			bitconv.cpix[0]=rows[y][xi++];
      			bitconv.cpix[1]=rows[y][xi++];
      			arr[1][y][x]=(float)bitconv.ipix;
      			bitconv.cpix[0]=rows[y][xi++];
      			bitconv.cpix[1]=rows[y][xi++];
      			arr[2][y][x]=(float)bitconv.ipix;
				*/
      			/* currently the (possible) alpha channel is ignored, i.e. the
      			   resulting image data looks like combined with black backgroung */
      			if(color_type & PNG_COLOR_MASK_ALPHA) xi+=2;
      		} else {
      			arr[0][y][x]=(float)rows[y][xi++];
      			arr[1][y][x]=(float)rows[y][xi++];
      			arr[2][y][x]=(float)rows[y][xi++];
      			/* currently the (possible) alpha channel is ignored, i.e. the
      			   resulting image data looks like combined with black backgroung */
      			if(color_type & PNG_COLOR_MASK_ALPHA) xi++;
      		}
      	}
      }

      for (y = 0; y < height; y++)
         png_free(png_ptr,rows[y]);
      png_free(png_ptr,rows);
   }

   /* read rest of file, and get additional chunks in info_ptr - REQUIRED */
   png_read_end(png_ptr, info_ptr);

   /* At this point you have read the entire image */
   /* clean up after the read, and free any memory allocated - REQUIRED */
   png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);

   fclose(fp);
   return 1;
} /* readPNGfile3D ------------------------------------------------------ */

/**************************************************************************/
int
writePNGfile(const char *filename,float ***arr,int planes, int nr, int nc, int oneplane, int sixteenbit)
{
   /*! \param 'planes' should be 1 for gray-scale or 3 for RGB */

   FILE *fp;
   png_structp png_ptr;
   png_infop info_ptr;
   /*png_colorp palette;*/
   png_color_8 sig_bit;
#ifdef PNG_TEXT_SUPPORTED
   png_text desc[4];
#endif
   png_bytep row;
   int x,y,xi;
   float **arr2=NULL;
   int fileplanes;
   int true_bit_depth;
   //   union {unsigned short ipix; char cpix[2];} bitconv;
   unsigned short byte1,byte2;

   if(sixteenbit) true_bit_depth=16; else true_bit_depth=8;
   if(planes==1) /* grayscale image */
   {
      if(oneplane<0) arr2=(float**)arr; else arr2 = arr[oneplane];
      fileplanes=1;
   } else fileplanes=3;

   if(planes<0) return IMGIO_WRONG_PARAMETER;
   /* if((planes!=1)&&(planes!=3)) return IMGIO_WRONG_PARAMETER; */
   /* \warning PS changed the behaviour so that for nonstandard planes an RGB image is stored
               containing simply the first 3 planes (2 in case of planes==2) */

   /* open the file */
   fp = fopen(filename, "wb");
   if (fp == NULL) return IMGIO_FILE_NOT_OPENED;

   /* Create and initialize the png_struct with the desired error handler
    * functions.  If you want to use the default stderr and longjump method,
    * you can supply NULL for the last three parameters.  We also check that
    * the library version is compatible with the one used at compile time,
    * in case we are using dynamically linked libraries.  REQUIRED.
    */
   png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, (png_voidp)NULL, NULL, NULL);
   if (png_ptr == NULL)
   {
      fclose(fp);
      return IMGIO_ALLOCATION_PROBLEM;
   }

   /* Allocate/initialize the image information data.  REQUIRED */
   info_ptr = png_create_info_struct(png_ptr);
   if (info_ptr == NULL)
   {
      fclose(fp);
      png_destroy_write_struct(&png_ptr,  png_infopp_NULL);
      return IMGIO_ALLOCATION_PROBLEM;
   }

   /* Set error handling.  REQUIRED if you aren't supplying your own
    * error handling functions in the png_create_write_struct() call.
    */
   if (setjmp(png_jmpbuf(png_ptr)))
   {
      /* If we get here, we had a problem reading the file */
      fclose(fp);
      png_destroy_write_struct(&png_ptr, &info_ptr);
      return IMGIO_FILE_WRITE_PROBLEM;
   }

   /* set up the output control if you are using standard C streams */
   png_init_io(png_ptr, fp);

   /* Set the image information here.  Width and height are up to 2^31,
    * bit_depth is one of 1, 2, 4, 8, or 16, but valid values also depend on
    * the color_type selected. color_type is one of PNG_COLOR_TYPE_GRAY,
    * PNG_COLOR_TYPE_GRAY_ALPHA, PNG_COLOR_TYPE_PALETTE, PNG_COLOR_TYPE_RGB,
    * or PNG_COLOR_TYPE_RGB_ALPHA.  interlace is either PNG_INTERLACE_NONE or
    * PNG_INTERLACE_ADAM7, and the compression_type and filter_type MUST
    * currently be PNG_COMPRESSION_TYPE_BASE and PNG_FILTER_TYPE_BASE. REQUIRED
    */
   if(planes==1)
   png_set_IHDR(png_ptr, info_ptr, nc, nr, true_bit_depth, PNG_COLOR_TYPE_GRAY,
      PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
   else
   png_set_IHDR(png_ptr, info_ptr, nc, nr, true_bit_depth, PNG_COLOR_TYPE_RGB,
      PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);


   /*! \warning - not finished ??.. */

   sig_bit.red = true_bit_depth; /*true_red_bit_depth;*/
   sig_bit.green = true_bit_depth; /*true_green_bit_depth;*/
   sig_bit.blue = true_bit_depth; /*true_blue_bit_depth;*/
   sig_bit.gray = true_bit_depth; /*true_gray_bit_depth;*/
   png_set_sBIT(png_ptr, info_ptr, &sig_bit);

   /* Optional gamma chunk is strongly suggested if you have any guess
    * as to the correct gamma of the image.
    */
   /*png_set_gAMA(png_ptr, info_ptr, gamma);*/

#ifdef PNG_TEXT_SUPPORTED
   desc[0].key = (char*)"Title";
   desc[0].text = (char*)filename;
   desc[0].compression = PNG_TEXT_COMPRESSION_NONE;
   desc[1].key = (char*)"Description";
   desc[1].text = (char*)"This file was generated using the VR (texture and pattern modelling) library.";
   desc[1].compression = PNG_TEXT_COMPRESSION_NONE;
   desc[2].key = (char*)"Copyright";
   desc[2].text = (char*)"Dr. Michal Haindl, Institute of Information Theory and Automation (UTIA), Academy of Sciences of the Czech Republic, Prague.";
   desc[2].compression = PNG_TEXT_COMPRESSION_NONE;
   desc[3].key = (char*)"Contact";
   desc[3].text = (char*)"noname@universe.com";
   desc[3].compression = PNG_TEXT_COMPRESSION_NONE;
#ifdef PNG_iTXt_SUPPORTED
   desc[0].lang = NULL;
   desc[1].lang = NULL;
   desc[2].lang = NULL;
   desc[3].lang = NULL;
#endif
   png_set_text(png_ptr, info_ptr, (png_textp)(&desc), 4);
#endif

   /* Write the file header information.  REQUIRED */
   png_write_info(png_ptr, info_ptr);

   /* Once we write out the header, the compression type on the text
    * chunks gets changed to PNG_TEXT_COMPRESSION_NONE_WR or
    * PNG_TEXT_COMPRESSION_zTXt_WR, so it doesn't get written out again
    * at the end.
    */


   /* The easiest way to write the image (you may have a different memory
    * layout, however, so choose what fits your needs best).  You need to
    * use the first method if you aren't handling interlacing yourself.
    */
   // ?????????????? pretypovani ok????
   if(sixteenbit) row = (png_bytep)png_malloc(png_ptr, (long)nc*(long)fileplanes*(long)2);
   else row = (png_bytep)png_malloc(png_ptr, (long)nc*(long)fileplanes);
   if (row==NULL)
   {
      /* Free all of the memory associated with the png_ptr and info_ptr */
      png_destroy_write_struct(&png_ptr, &info_ptr);
      fclose(fp);
      return IMGIO_ALLOCATION_PROBLEM;
   }
   /* Write 1 row at a time. */
   for (y = 0; y < nr; y++)
   {
      if(planes==1)
      {
      	  for(x=0;x<nc;x++) if(sixteenbit)
      	  {
            byte1 = byte2 = (unsigned short)arr2[y][x];
			byte1 >>= 8;
            byte2 <<= 8;
            byte2 >>= 8;
			row[2*x] = (unsigned char)(byte2);
			row[2*x+1] = (unsigned char)(byte1);
/*     			
      	  	bitconv.ipix=(unsigned short)arr2[y][x];
      	  	row[2*x]=bitconv.cpix[0];
      	  	row[2*x+1]=bitconv.cpix[1];
*/
      	  } else row[x]=(unsigned char)arr2[y][x];
      }
      else
      {
      	  xi=0;
      	  for(x=0;x<nc;x++) if(sixteenbit)
      	  {
            byte1 = byte2 = (unsigned short)arr[0][y][x];
			byte1 >>= 8;
            byte2 <<= 8;
            byte2 >>= 8;
			row[xi++] = (unsigned char)(byte2);
			row[xi++] = (unsigned char)(byte1);
            byte1 = byte2 = (unsigned short)arr[1][y][x];
			byte1 >>= 8;
            byte2 <<= 8;
            byte2 >>= 8;
			row[xi++] = (unsigned char)(byte2);
			row[xi++] = (unsigned char)(byte1);

            /*
      	  	bitconv.ipix=(unsigned short)arr[0][y][x];
      	  	row[xi++]=bitconv.cpix[0];
      	  	row[xi++]=bitconv.cpix[1];
      	  	bitconv.ipix=(unsigned short)arr[1][y][x];
      	  	row[xi++]=bitconv.cpix[0];
      	  	row[xi++]=bitconv.cpix[1];
            */
      	  	if(planes>=3)
      	  	{
            byte1 = byte2 = (unsigned short)arr[2][y][x];
			byte1 >>= 8;
            byte2 <<= 8;
            byte2 >>= 8;
			row[xi++] = (unsigned char)(byte2);
			row[xi++] = (unsigned char)(byte1);
              /*
      	  		bitconv.ipix=(unsigned short)arr[2][y][x];
      	  		row[xi++]=bitconv.cpix[0];
      	  		row[xi++]=bitconv.cpix[1];
              */
      	  	} else {row[xi++]=255; row[xi++]=255;}
      	  } else {
      	  	row[xi++]=(unsigned char)arr[0][y][x];
      	  	row[xi++]=(unsigned char)arr[1][y][x];
      	  	if(planes>=3) row[xi++]=(unsigned char)arr[2][y][x]; else row[xi++]=255;
      	  }
      }
      png_write_rows(png_ptr, &row, 1);
   }
   png_free(png_ptr,row);

   /* You can write optional chunks like tEXt, zTXt, and tIME at the end
    * as well.  Shouldn't be necessary in 1.1.0 and up as all the public
    * chunks are supported and you can use png_set_unknown_chunks() to
    * register unknown chunks into the info structure to be written out.
    */

   /* It is REQUIRED to call this to finish writing the rest of the file */
   png_write_end(png_ptr, info_ptr);


   /*! \warning - to be checked ... */
   /* Similarly, if you png_malloced any data that you passed in with
      png_set_something(), such as a hist or trans array, free it here,
      when you can be sure that libpng is through with it. */
   /*png_free(png_ptr, trans);*/
   /*trans=NULL;*/

   /* clean up after the write, and free any memory allocated */
   png_destroy_write_struct(&png_ptr, &info_ptr);

   /* close the file */
   fclose(fp);
   return 1;
} /* writePNGfile ---------------------------------------------------- */

int
writePNGfile1D(const char *filename, unsigned char *arr,int planes, int nr, int nc, int oneplane)
{
   /*! \param 'planes' should be 1 for gray-scale or 3 for RGB */

   FILE *fp;
   png_structp png_ptr;
   png_infop info_ptr;
   /*png_colorp palette;*/
   png_color_8 sig_bit;
#ifdef PNG_TEXT_SUPPORTED
   png_text desc[4];
#endif
   png_bytep row;
   int x,y,xi;
   float **arr2=NULL;
   int fileplanes;
   int true_bit_depth;
   union {unsigned int ipix; char cpix[2];} bitconv;

   true_bit_depth=8;

   if(planes==1) /* grayscale image */
     fileplanes=1;
   else 
     fileplanes=3;

   if(planes<0) return IMGIO_WRONG_PARAMETER;
   /* if((planes!=1)&&(planes!=3)) return IMGIO_WRONG_PARAMETER; */
   /* \warning PS changed the behaviour so that for nonstandard planes an RGB image is stored
               containing simply the first 3 planes (2 in case of planes==2) */

   /* open the file */
   fp = fopen(filename, "wb");
   if (fp == NULL) return IMGIO_FILE_NOT_OPENED;

   /* Create and initialize the png_struct with the desired error handler
    * functions.  If you want to use the default stderr and longjump method,
    * you can supply NULL for the last three parameters.  We also check that
    * the library version is compatible with the one used at compile time,
    * in case we are using dynamically linked libraries.  REQUIRED.
    */
   png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, (png_voidp)NULL, NULL, NULL);
   if (png_ptr == NULL)
   {
      fclose(fp);
      return IMGIO_ALLOCATION_PROBLEM;
   }

   /* Allocate/initialize the image information data.  REQUIRED */
   info_ptr = png_create_info_struct(png_ptr);
   if (info_ptr == NULL)
   {
      fclose(fp);
      png_destroy_write_struct(&png_ptr,  png_infopp_NULL);
      return IMGIO_ALLOCATION_PROBLEM;
   }

   /* Set error handling.  REQUIRED if you aren't supplying your own
    * error handling functions in the png_create_write_struct() call.
    */
   if (setjmp(png_jmpbuf(png_ptr)))
   {
      /* If we get here, we had a problem reading the file */
      fclose(fp);
      png_destroy_write_struct(&png_ptr, &info_ptr);
      return IMGIO_FILE_WRITE_PROBLEM;
   }

   /* set up the output control if you are using standard C streams */
   png_init_io(png_ptr, fp);

   /* Set the image information here.  Width and height are up to 2^31,
    * bit_depth is one of 1, 2, 4, 8, or 16, but valid values also depend on
    * the color_type selected. color_type is one of PNG_COLOR_TYPE_GRAY,
    * PNG_COLOR_TYPE_GRAY_ALPHA, PNG_COLOR_TYPE_PALETTE, PNG_COLOR_TYPE_RGB,
    * or PNG_COLOR_TYPE_RGB_ALPHA.  interlace is either PNG_INTERLACE_NONE or
    * PNG_INTERLACE_ADAM7, and the compression_type and filter_type MUST
    * currently be PNG_COMPRESSION_TYPE_BASE and PNG_FILTER_TYPE_BASE. REQUIRED
    */
   if(planes==1)
   png_set_IHDR(png_ptr, info_ptr, nc, nr, true_bit_depth, PNG_COLOR_TYPE_GRAY,
      PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
   else
   png_set_IHDR(png_ptr, info_ptr, nc, nr, true_bit_depth, PNG_COLOR_TYPE_RGB,
      PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

   /*! \warning - not finished ??.. */

   sig_bit.red = true_bit_depth; /*true_red_bit_depth;*/
   sig_bit.green = true_bit_depth; /*true_green_bit_depth;*/
   sig_bit.blue = true_bit_depth; /*true_blue_bit_depth;*/
   sig_bit.gray = true_bit_depth; /*true_gray_bit_depth;*/
   png_set_sBIT(png_ptr, info_ptr, &sig_bit);

   /* Optional gamma chunk is strongly suggested if you have any guess
    * as to the correct gamma of the image.
    */
   /*png_set_gAMA(png_ptr, info_ptr, gamma);*/

#ifdef PNG_TEXT_SUPPORTED
   desc[0].key = (char*)"Title";
   desc[0].text = (char*)filename;
   desc[0].compression = PNG_TEXT_COMPRESSION_NONE;
   desc[1].key = (char*)"Description";
   desc[1].text = (char*)"This file was generated using the VR (texture and pattern modelling) library.";
   desc[1].compression = PNG_TEXT_COMPRESSION_NONE;
   desc[2].key = (char*)"Copyright";
   desc[2].text = (char*)"Dr. Michal Haindl, Institute of Information Theory and Automation (UTIA), Academy of Sciences of the Czech Republic, Prague.";
   desc[2].compression = PNG_TEXT_COMPRESSION_NONE;
   desc[3].key = (char*)"Contact";
   desc[3].text = (char*)"haindl@utia.cas.cz";
   desc[3].compression = PNG_TEXT_COMPRESSION_NONE;
#ifdef PNG_iTXt_SUPPORTED
   desc[0].lang = NULL;
   desc[1].lang = NULL;
   desc[2].lang = NULL;
   desc[3].lang = NULL;
#endif
   png_set_text(png_ptr, info_ptr, (png_textp)(&desc), 4);
#endif

   /* Write the file header information.  REQUIRED */
   png_write_info(png_ptr, info_ptr);

   /* Once we write out the header, the compression type on the text
    * chunks gets changed to PNG_TEXT_COMPRESSION_NONE_WR or
    * PNG_TEXT_COMPRESSION_zTXt_WR, so it doesn't get written out again
    * at the end.
    */


   /* The easiest way to write the image (you may have a different memory
    * layout, however, so choose what fits your needs best).  You need to
    * use the first method if you aren't handling interlacing yourself.
    */
   // ?????????????? pretypovani ok????
   row = (png_bytep)png_malloc(png_ptr, (long)nc*(long)fileplanes);
   if (row==NULL)
   {
      /* Free all of the memory associated with the png_ptr and info_ptr */
      png_destroy_write_struct(&png_ptr, &info_ptr);
      fclose(fp);
      return IMGIO_ALLOCATION_PROBLEM;
   }
   /* Write 1 row at a time. */

   // horizontalne prevraceny obraz
   int gi=4*nr*nc-1;
   for (y = 0; y < nr; y++)
   {
     if(planes==1)
       {
         for(x=0;x<nc;x++) 
           row[x] = arr[gi++];
       }
     else
       {
         xi = 3*nc-1;
         for(x=0;x<nc;x++)
           {
             gi--;
             row[xi--] = arr[gi--];
             row[xi--] = arr[gi--];
             row[xi--] = arr[gi--]; 
           }
       }

   /*
   int gi=0;
   for (y = 0; y < nr; y++)
   {
     if(planes==1)
       {
         for(x=0;x<nc;x++) 
           row[x] = arr[gi++];
       }
     else
       {
         xi = 0;
         for(x=0;x<nc;x++)
           {
             row[xi++] = arr[gi++];
             row[xi++] = arr[gi++];
             row[xi++] = arr[gi++];
             gi++;
           }
       }
   */
     png_write_rows(png_ptr, &row, 1);
   }
   png_free(png_ptr,row);

   /* You can write optional chunks like tEXt, zTXt, and tIME at the end
    * as well.  Shouldn't be necessary in 1.1.0 and up as all the public
    * chunks are supported and you can use png_set_unknown_chunks() to
    * register unknown chunks into the info structure to be written out.
    */

   /* It is REQUIRED to call this to finish writing the rest of the file */
   png_write_end(png_ptr, info_ptr);


   /*! \warning - to be checked ... */
   /* Similarly, if you png_malloced any data that you passed in with
      png_set_something(), such as a hist or trans array, free it here,
      when you can be sure that libpng is through with it. */
   /*png_free(png_ptr, trans);*/
   /*trans=NULL;*/

   /* clean up after the write, and free any memory allocated */
   png_destroy_write_struct(&png_ptr, &info_ptr);

   /* close the file */
   fclose(fp);
   return 1;
} /* writePNGfile1D ---------------------------------------------------- */

