// ================================================================
// $Id: $
//
// render0.cpp
//
//     This implements one-point-light illumination
//
// Initial coding by Vlastimil Havran, 2013.

#include <cassert>
#include <cmath>

#include "configg.h"
#include "AuxFuncs.h"
#include "lightenv.h"
#include "brdfmodel.h"
#include "vrmlsup.h"
#include "tangles.h"
#include "render.h"

// global variables
// the debug counters on the success of BRDF sampling
long long int failBRDFcnt = 0L;
long long int evalBRDFcnt = 0L;
long long int failEMcnt = 0L;
long long int evalEMcnt = 0L;
// counters for a single pixel
int evalEMcntST = 0;
int evalBRDFcntST = 0;
float evalValueVIS = 0;

// ================================================================================
// Evaluating a single point light of white colour
void         // interpolated vertices, tangents, normals and binormals
computeMode0(float Vint[3], float Tint[3], float Nint[3], float Bint[3], 
	     float t_v, float p_v, int irow, int jcol, int BTFindex, float normF)
{
  float t_i, p_i;
  float result[3];
  // the final value for the pixel
  result[0] = result[1] = result[2] = 0.f;
  // compute the direction of the light for environment map here
  TA->computeIlluAngles(Vint, Tint, Bint, Nint, &t_i, &p_i);
  assert(t_i >= 0.f); assert(t_i <= M_PI);
  
  if (t_i < M_PI/2.0f) {
    float brdfRGB[3];
    // t_i is in radians
    // here we get the common BRDF*cos(theta) value irrespective of the
    // underlying BRDF model
    bool tt = brdf->getBRDFValue(irow, jcol, t_v, p_v, t_i, p_i, brdfRGB);
    const float cosTheta = cos(t_i) + 0.0001f;
    for(int isp=0;isp<=2;isp++) {
      result[isp] = brdfRGB[isp]; // use BRDF/BTF model only, intensity of light is one
    } // for all channels
  } // if

  // copy the data to the output texture
  for(int isp=0;isp<=2;isp++) {
    BTFimageF[BTFindex+isp] = normF * result[isp];
  }  
  setCountersST(0, 1);
  return;
}

